# Generated from stlParser.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .stlParser import stlParser
else:
    from stlParser import stlParser

# This class defines a complete listener for a parse tree produced by stlParser.
class stlParserListener(ParseTreeListener):

    # Enter a parse tree produced by stlParser#stlSpecification.
    def enterStlSpecification(self, ctx:stlParser.StlSpecificationContext):
        pass

    # Exit a parse tree produced by stlParser#stlSpecification.
    def exitStlSpecification(self, ctx:stlParser.StlSpecificationContext):
        pass


    # Enter a parse tree produced by stlParser#predicateExpr.
    def enterPredicateExpr(self, ctx:stlParser.PredicateExprContext):
        pass

    # Exit a parse tree produced by stlParser#predicateExpr.
    def exitPredicateExpr(self, ctx:stlParser.PredicateExprContext):
        pass


    # Enter a parse tree produced by stlParser#opFutureExpr.
    def enterOpFutureExpr(self, ctx:stlParser.OpFutureExprContext):
        pass

    # Exit a parse tree produced by stlParser#opFutureExpr.
    def exitOpFutureExpr(self, ctx:stlParser.OpFutureExprContext):
        pass


    # Enter a parse tree produced by stlParser#parenPhiExpr.
    def enterParenPhiExpr(self, ctx:stlParser.ParenPhiExprContext):
        pass

    # Exit a parse tree produced by stlParser#parenPhiExpr.
    def exitParenPhiExpr(self, ctx:stlParser.ParenPhiExprContext):
        pass


    # Enter a parse tree produced by stlParser#opUntilExpr.
    def enterOpUntilExpr(self, ctx:stlParser.OpUntilExprContext):
        pass

    # Exit a parse tree produced by stlParser#opUntilExpr.
    def exitOpUntilExpr(self, ctx:stlParser.OpUntilExprContext):
        pass


    # Enter a parse tree produced by stlParser#opGloballyExpr.
    def enterOpGloballyExpr(self, ctx:stlParser.OpGloballyExprContext):
        pass

    # Exit a parse tree produced by stlParser#opGloballyExpr.
    def exitOpGloballyExpr(self, ctx:stlParser.OpGloballyExprContext):
        pass


    # Enter a parse tree produced by stlParser#opLogicalExpr.
    def enterOpLogicalExpr(self, ctx:stlParser.OpLogicalExprContext):
        pass

    # Exit a parse tree produced by stlParser#opLogicalExpr.
    def exitOpLogicalExpr(self, ctx:stlParser.OpLogicalExprContext):
        pass


    # Enter a parse tree produced by stlParser#opReleaseExpr.
    def enterOpReleaseExpr(self, ctx:stlParser.OpReleaseExprContext):
        pass

    # Exit a parse tree produced by stlParser#opReleaseExpr.
    def exitOpReleaseExpr(self, ctx:stlParser.OpReleaseExprContext):
        pass


    # Enter a parse tree produced by stlParser#opNextExpr.
    def enterOpNextExpr(self, ctx:stlParser.OpNextExprContext):
        pass

    # Exit a parse tree produced by stlParser#opNextExpr.
    def exitOpNextExpr(self, ctx:stlParser.OpNextExprContext):
        pass


    # Enter a parse tree produced by stlParser#opPropExpr.
    def enterOpPropExpr(self, ctx:stlParser.OpPropExprContext):
        pass

    # Exit a parse tree produced by stlParser#opPropExpr.
    def exitOpPropExpr(self, ctx:stlParser.OpPropExprContext):
        pass


    # Enter a parse tree produced by stlParser#opNegExpr.
    def enterOpNegExpr(self, ctx:stlParser.OpNegExprContext):
        pass

    # Exit a parse tree produced by stlParser#opNegExpr.
    def exitOpNegExpr(self, ctx:stlParser.OpNegExprContext):
        pass


    # Enter a parse tree produced by stlParser#predicate.
    def enterPredicate(self, ctx:stlParser.PredicateContext):
        pass

    # Exit a parse tree produced by stlParser#predicate.
    def exitPredicate(self, ctx:stlParser.PredicateContext):
        pass


    # Enter a parse tree produced by stlParser#interval.
    def enterInterval(self, ctx:stlParser.IntervalContext):
        pass

    # Exit a parse tree produced by stlParser#interval.
    def exitInterval(self, ctx:stlParser.IntervalContext):
        pass


